<?php

namespace KalmoyaElementalCloner;

defined('C5_EXECUTE') or exit('Access Denied.');

use Concrete\Core\Application\Application;

class Queue
{
    protected $data = [];
    protected $app;
    protected $dataCache;
    protected $cacheEngine;

    public function __construct(Application $app, $handle)
    {
        $this->app = $app;
        $this->cacheEngine = $this->app->make('cache/expensive');
        $this->dataCache = $this->cacheEngine->getItem($handle);

        // We check that the item already exists and is not expired
        if (!$this->dataCache->isMiss()) {
            // Everything's fine so we grab our data from the cache
            $this->data = $this->dataCache->get();
        }
    }

    public function send($data, $save = false)
    {
        $this->data[] = $data;
        if ($save) {
            $this->saveQueue();
        }
    }

    public function saveQueue()
    {
        $this->dataCache->lock();

        // We set the data and give it an expiry delay of an hour in seconds.
        $this->dataCache->set($this->data)->expiresAfter(3600);

        // Finally we save our data to the cache
        $this->cacheEngine->save($this->dataCache);
    }

    public function deleteQueue()
    {
        // empty cache and set this->data to []
        $this->data = [];
        if (!$this->dataCache->isMiss()) {
            $this->dataCache->clear();
        }
    }

    public function receive($limit = 1)
    {
        return array_slice($this->data, 0, $limit);
    }

    public function count()
    {
        return count($this->data);
    }

    public function deleteMessage($msg)
    {
        $index = array_search($msg, $this->data);
        if (false !== $index) {
            array_splice($this->data, $index, 1);
            $this->saveQueue();
        }
    }
}
